/* */
#include <avr/interrupt.h>
#include "reset.h"
#include "globals.h"
#include "meter.h"
#include "led.h"
#include "globals.h"


// the noinit is important, as this value is read following a reset when
// setting up the min/max meter range.
static volatile uint16_t _meter_lastvalue __attribute__ ((section (".noinit"))) ;

uint16_t inline meter_lastvalue(void){ return( _meter_lastvalue ) ; }

void meter_init( uint8_t source ){
	// set up the ADC input...
	switch (source){
		case SOURCE_Temperature :
			ADMUX = (0<<REFS2)|(1<<REFS1)|(0<<REFS0)|(15 << MUX0) ; // use internal temp sensor
			break ;
		case SOURCE_SensePin : // drop through to default
		default:
			ADMUX = (0<<REFS2)|(0<<REFS1)|(0<<REFS0)|(1 << MUX0) ; // 0->vcc on B2 (pin7)
	}
	
	// ADSCRA: Enable AD, Start Conversion, AutoTrigger enable, Enable Interrupt, /64 clock scaling.
	ADCSRA = (1<<ADEN)|(1<<ADSC)|(1<<ADATE)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(0<<ADPS0) ;

	// ADSCRB: Free Running
	ADCSRB = (0<<ADTS2)|(0<<ADTS1)|(0<<ADTS0) ;
	
	// mark the sense pin as input...
	DDRB &= ~SENSE_bm ;
	
	// should I be disabling the digital input on ADC2?
}

ISR( ADC_vect ){
	// an ADC conversion has been completed.
	_meter_lastvalue = ADC ;
}

void _LEDFill( uint16_t start, uint16_t end ){
	// start and end are binary fractions of LEDs.
	// start MUST be <= end
	
	// Essentially, the top byte is the LED, and the
	// bottom byte is the fraction of the LED

	for( uint8_t ledindex = 0 ; ledindex < 8 ; ledindex++ ){
		uint16_t LEDstart = (ledindex + g_myfirstled) << 8 ;
		uint16_t LEDend = LEDstart | 0xff ;
		uint16_t fillstart= (LEDstart > start)? LEDstart : start ;
		uint16_t fillend  = (LEDend < end)? LEDend : end ;
		if (fillend < fillstart){
			//nothing is filled in this LED.
			g_led_target[ ledindex ] = 0 ;
		} else {
			g_led_target[ ledindex ] = fillend - fillstart ;
		}
	}
}

void __attribute__((noreturn)) _fn_genericbar( const uint16_t min, const uint16_t max ){
// Calculate a constant for use later...
int16_t MaxLED = (int16_t)g_chain_ledlength * (int16_t)256 ;
const int32_t coeff = ((int32_t)MaxLED * (int32_t)256) / ((int32_t)max - (int32_t)min) ;
uint16_t bound_low = (min < max)?min:max ;
uint16_t bound_high = (min < max)?max:min ;
// 2. low = 0, high = 1023

	while(1){
		while(g_tick==0){}
		g_tick = 0 ;
		// now convert the 0..1024 value in _meter_lastvalue to a meter piccy.
		int16_t value = _meter_lastvalue ;
		if (value < bound_low) value = bound_low ;
		if (value > bound_high) value = bound_high ;
		int32_t result = (value - (int16_t) min) * coeff ;
		uint16_t end = ((result & 0xff0000) >> 8) | ((result & 0xff00) >> 8) ;
		
		_LEDFill( 0, end ) ;
	}
}

void __attribute__((noreturn)) _fn_genericdot( const uint16_t min, const uint16_t max ){
// Calculate a constant for use later...
int16_t MaxLED = (int16_t)(g_chain_ledlength - 1) * (int16_t)256 ;
const int32_t coeff = ((int32_t)MaxLED * (int32_t)256) / ((int32_t)max - (int32_t)min) ;
uint16_t bound_low = (min < max)?min:max ;
uint16_t bound_high = (min < max)?max:min ;
// 2. low = 0, high = 1023

	while(1){
		while(g_tick==0){}
		g_tick = 0 ;
		// now convert the 0..1024 value in _meter_lastvalue to a meter piccy.
		int16_t value = _meter_lastvalue ;
		if (value < bound_low) value = bound_low ;
		if (value > bound_high) value = bound_high ;
		int32_t result = (value - (int16_t) min) * coeff ;
		uint16_t end = ((result & 0xff0000) >> 8) | ((result & 0xff00) >> 8) ;
		
		_LEDFill( end, end + 255 ) ;
	}
}

/*
void __attribute__((noreturn)) _fn_genericmeter( uint8_t source, uint8_t data ){
	meter_init( source ) ;
	sei();
	while(1){
		while(g_tick==0){}
		g_tick = 0 ;
		// now convert the 0..1024 value in _meter_lastvalue to a meter piccy.
		
		uint16_t value = _meter_lastvalue ;
		
		if (source == SOURCE_Temperature){
			// scaling for temp.
			value -= 290 ;
			value *= 40 ;
		}
		
		uint8_t h = (value & 0xff00) >> 8 ;
		uint8_t l = (value & 0xff) ;
		uint8_t led = h << 1 ;
		if (l & 0b10000000) led+=1 ;
		uint8_t bright = l & 0b01111111 ;
		bright <<= 1 ;
		
		for( uint8_t ledindex = 0 ; ledindex < 8 ; ledindex++ ){
			if ( led > ledindex ) {
				g_led_target[ ledindex ] = 255 ;
			}	else {
				if (led == ledindex){
					g_led_target[ ledindex ] = bright ;
				}	else	{
					g_led_target[ ledindex ] = 0 ;
				}
			}
		}
	}
}
*/

void __attribute__((noreturn)) _generic_meter( uint8_t data, uint8_t source, uint16_t min, uint16_t max ){
	// firstly store the new style if the device was reset using the
	// button...
	if (g_reset_reason & (1<<EXTRF)){
		g_function_data = data ;
		eeprom_save_state();
	}
	// now increment the data byte, so the next style will be used
	// following a button press...
	uint8_t nextdata = data + 1 ;
	if (nextdata > 1) nextdata = 0 ;
	g_function_data = nextdata ;

	// set the fastest LED effect.
	// the 'target' intensity is jumped to immediately.
	// it gives the fastest possible response.
	led_seteffect( STYLE_Direct ) ;
	
	// start off the ADC stuff...
	meter_init( source ) ;
	sei();

	// now call the correct style of meter...
	switch( data ){
		case 1:
			_fn_genericdot( min, max ) ;
			// no break necessary - the above function doesn't return.
		default:
			_fn_genericbar( min, max ) ;
			// no break necessary - the above function doesn't return.
	}
}

void __attribute__((noreturn)) fn_thermometer( uint8_t data ){
	_generic_meter( data, SOURCE_Temperature, 500, 1023 ) ;
}

void __attribute__((noreturn)) fn_meter( uint8_t data ){
	_generic_meter( data, SOURCE_SensePin, g_meter_min, g_meter_max ) ;
}
