/*
*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include "led.h"
#include "globals.h"

/*
See transmit.c for details of the data signalling.

A falling edge on the sense line indicates the start of a bit.

If a falling edge is found after 256us or more from the previous edge,
then it's the first bit in a data transmission.
Timer1's overflow is used to find this.

Each message is around 1792us long.
Any data received is stored in a small buffer.
This allows for some latency in the processing of the incoming data.
*/

#ifdef DEBUG
#define RXDEBUG
#endif

void _buffer_init( void ) ;
uint8_t _buffer_isempty( void ) ;
uint8_t _buffer_write( uint16_t data ) ;
uint16_t _buffer_read( void ) ;

ISR( INT0_vect ){
static uint8_t s_data ;
static uint8_t s_data_hi ;
static uint8_t s_bitcount = 0 ;
// we've just found a falling edge...

#ifdef RXDEBUG
	PORTB |= (1<<PB0) ; // set pin 0 ON
#endif


	if (TIFR & ( 1 << TOV1 )){ // has there been a timer overflow?
		s_bitcount = 0 ; // yes: therefore this is a start bit
	}
	TCNT1 = 0 ; // reset the timer
	TIFR = (1<<TOV1) ; // clear the timer flag

	s_data <<= 1 ;
	// the sense pin will hold the bit being transmitted... Shift it in...
	if (PINB & SENSE_bm) s_data |= 0b1 ;
	s_bitcount++ ;
	if (s_bitcount == 8){
		// we've received half of the data...
		s_data_hi = s_data ;
	}
	if (s_bitcount == 16){
		// the data transfer is complete.
		_buffer_write( s_data_hi << 8 | s_data ) ;
		s_bitcount = 0 ;
	}
#ifdef RXDEBUG
	PORTB &= ~(1<<PB0) ; // set pin 0 OFF
#endif
}

void rx_init( void ){
	// Configure Timer1 so that it will overflow in 256us.
	// At 8Mhz, that's 2048 clocks, hence use the /8 prescaler.
	TCCR1 |= (0b0100 << CS10) ;
	TCNT1 = 0 ;
	TIFR = (1<<TOV1) ; // clear the timer overflow flag
	
	// Configure INT0 (PB2) to raise an interrupt on a falling edge.
	MCUCR |= (0b10 << ISC00) ;
	GIMSK |= (1 << INT0) ;

	_buffer_init() ;
}

uint8_t rx_ready( void ){
	return(!_buffer_isempty()) ;
}

// note that the following routine blocks until data is available.
uint16_t rx_getdata( void ){
	while( _buffer_isempty()) ; // wait until the buffer's got something.
	return( _buffer_read() ) ;
}

void rx_insertdata( uint16_t data ){
	_buffer_write( data ) ;
}

// BUF_SIZE can be from 2 to 255
#define BUF_SIZE 10

static uint16_t m_buffer[ BUF_SIZE ] ; // RX buffer
static uint8_t m_readindex ;
static uint8_t m_writeindex ;

void _buffer_init( void ){
	m_readindex = 0 ;
	m_writeindex = 0 ;
}

uint8_t _buffer_isempty( void ){
	return( m_readindex == m_writeindex ) ;
}

uint8_t _buffer_write( uint16_t data ){
	// returns 0 for success.
	uint8_t temp = m_writeindex + 1 ;
	if (temp >= BUF_SIZE) temp = 0 ;
	uint8_t is_buffer_full = (temp==m_readindex) ;
	if (!is_buffer_full){
		m_buffer[ m_writeindex ] = data ;
		m_writeindex = temp ;
	}
	return is_buffer_full ;
}

uint16_t _buffer_read( void ){
	uint16_t data = m_buffer[ m_readindex ] ;
	uint8_t temp = m_readindex + 1 ;
	if (temp >= BUF_SIZE) temp = 0 ;
	m_readindex = temp ;
	return data ;
}
