/*
*/
#include <avr/eeprom.h>
#include "globals.h"
#include "serial.h"

#define READ_SERIAL_PIN (PINB & SENSE_bm)
#define SERIAL_PIN_MARK READ_SERIAL_PIN

// How long is the break in serial communications?
// (measured in bit widths)
#define BREAK_MIN_DURATION 20

#define BIT_PERIOD_INVALID 0xffff
#define BIT_PERIOD_MAX 0xfffe

uint16_t m_bitperiod = BIT_PERIOD_INVALID; // duration in "delay loops"


/*

Serial data stream...

  | | | | | | | | | | | |  <- bit periods...
  
__   _________________        _ MARK LEVEL (TTL HIGH), Logical 0
  |_|_|_|_|_|_|_|_|_| |_      _ SPACE LEVEL (TTL LOW), Logical 1

   ^ 0 1 2 3 4 5 6 7 ^ ^
   |                 | |
   |                 | next byte's start bit
   'start        stop'

The stream of data for the SweeperMeter will be continually repeated.
To allow the SweeperMeter to easily identify the _start_ of the stream,
there will be a break of at least 40 bit periods before the first byte.
During this break, the line will idle at MARK (TTL HIGH).

To determine if a BREAK has occurred whilst looking for the next byte,
TIMER1 will be configured to tick such that a BREAK delay will cause
OCR1A to be hit.


*/


void serial_init( void ){
	// Configure the sense pin as input, with pullup...
	DDRB &= ~SENSE_bm ;
	PORTB |= SENSE_bm ;
	// and that's about it... everything from here-on-in is just software!

	// set up timer 0 to blink LEDs 7 and 8 upon serial data transitions
	// LED 7 is on when PB0 is OUTPUT and 1
	// LED 8 is on when PB0 is OUTPUT and 0
	// PB0 is OC0A
	// Sense is T0, hence can be used to clock Timer0.
	// When OCR0A is hit, PB0 will be toggled.
	// Timer0 will be in CTC mode.
	
	// toggle OC0A on compare match, CTC mode
	TCCR0A |= (0b01 << COM0A0)|(0b10 << WGM00) ;
	TCCR0B |= (0b110 << CS00) ; // use T0 as clock on falling edge.
	OCR0A = 1 ; // toggle after every falling edge.
	TCNT0 = 0 ;
	DDRB |= (1<<PB0) ; // set PB0 to output

}

void serial_finish( void ){
	// stop twinkling the LEDs...
	TCCR0A &= ~(0b11 << COM0A0) ; // don't mess with OC0A.
	TCCR0B &= ~(0b111 << CS00) ; // stop clocking T0
	DDRB &= ~(1<<PB0) ; // set PB0 to input
	PORTB &= ~(1<<PB0) ; // disable the pull-up resistor.
}


// the result of this function is stored in a module level variable, m_bitperiod.
void serial_autobaud( void )
{
// measure lots of bit transitions and find the shortest one.
// Note that there are no timeouts here. If nothing's happening on
// the sense pin, this routine will never return!

uint8_t ReadingsCount = 100 ; // measure 100 transitions...
uint16_t SmallestBitPeriod = BIT_PERIOD_INVALID ;
uint16_t BitPeriod ;


	// set up timer 1 to blink LEDs 1 and 2 at a regular interval
	// LED 1 is on when PB4 is OUTPUT and 1
	// LED 2 is on when PB4 is OUTPUT and 0
	// PB4 is OC1B, so with a slow timer, we can toggle the pin to blink
	// these LEDs in the background.
	TCCR1 |= (0b1111 << CS10) ;
	OCR1B = 0 ;
	GTCCR |= (0b01 << COM1B0) ; // toggle PB4 (OC1B) on compare match B.
	DDRB |= (1<<PB4) ; // set PB4 to output (enables LED1/LED2)
	// Note that the blinking LEDs 1 and 2 will be disabled after
	// autobauding because timer 1 is then re-purposed for use in
	// the get_byte function.



	// the pin must have been read as low (and, given there is an (internal) pull-up resistor on the pin) means there must be activity.
	// so.... let's go and read some widths of bits...
	while ( ReadingsCount != 0 )
	{
		BitPeriod = 0 ;

		if (SERIAL_PIN_MARK)
		{
			// the bit was a 1...
			// That means that I'm part-way through a MARK.
			// wait for the first SPACE...
			while(SERIAL_PIN_MARK);
			
			// START OF INNER LOOP (1)
			while(!SERIAL_PIN_MARK)
			{
				BitPeriod++ ;
				if (BitPeriod == 0) break ; // bitperiod rolled over = timeout
			}
			// END OF INNER LOOP (1)
		}
		else
		{
			// the bit was a 0...
			// that means I'm part-way through a SPACE.
			// wait for the first MARK
			while( !SERIAL_PIN_MARK) ;
			
			// START OF INNER LOOP (0)
			while( SERIAL_PIN_MARK )
			{
				BitPeriod++ ;
				if (BitPeriod == 0) break ; // bitperiod rolled over = timeout
			}
			// END OF INNER LOOP (0)
		}

		if (BitPeriod)
		{
			// this was a valid reading...
			ReadingsCount-- ;
			if ( BitPeriod < SmallestBitPeriod ){
				// this is the shortest bitperiod yet!
				SmallestBitPeriod = BitPeriod ;
			}
		}
	}
	// OK, I've now looked at enough bits!
	m_bitperiod = SmallestBitPeriod ;

	// re-configure timer 1. It used to flipflop LED 1/2 to indicate
	// autobauding. It will now be used to time 'getbyte'...
	TCCR1 &= ~(0b1111 << CS10) ; // turn off timer 1.
	DDRB &= ~(1 << PB4) ; // switch PB4 to input
	PORTB &= ~(1 << PB4) ; // turn off pullup to PB4

	// Timer1 needs to be configured such that after around 16 bit periods
	// a compare match point will be hit.
	// This can then be used to determine if a byte had a very long
	// BREAK before it was received, and is therefore the first byte
	// in the looping stream.
	uint8_t prescaler = 8 ; // prescaler of 8 is /128 (9 is /256 etc)
	uint16_t a = m_bitperiod ;
	while (a & 0xff00){
    a >>= 1 ;
    prescaler += 1 ;
	}
	// the low byte of a is the timer compare value.
	// prescaler gives the timer's prescaler.
	OCR1A = a & 0xff ; // if OCR1A is hit, it took > 16 bit periods.
	TCCR1 |= (prescaler << CS10) ;
}

// bitperioddelay is as a define, because if it's as an inline function, the (never actually called) function is still created.
#define bitperioddelay( d )\
do{\
	uint16_t i = d ;\
	while( i-- ){ asm volatile("nop; \n nop; "); };\
}while(0)

// This retrieves an 8-bit character.
// it is assumed 8-bit + 1 parity bit (odd)
uint8_t serial_getbyte( uint8_t *status )
{
	uint8_t c = 0;
	uint8_t BitCount = 0;
	const uint16_t BitPeriod_SlightlyShort = m_bitperiod - 2  ;
	const uint16_t HalfBitPeriod = (m_bitperiod-1) >> 1 ;
	const uint16_t OneAndAHalfBitPeriod = m_bitperiod + HalfBitPeriod ;
	
	// re-start timer 1 from 0.
	TCNT1 = 0 ;
	GTCCR |= (1<<PSR1) ; // re-start the prescaler
	TIFR = (1<<OCF1A) ; // clear output compare flag 1A
	
	// note that we're probably already within a stop-bit, or a break, both of which are MARK.
	while( !SERIAL_PIN_MARK ); // wait for mark
	while( SERIAL_PIN_MARK ); // wait for space
	
	// we're now right at the beginning of the start bit.
	// wait until we're in the middle of the first bit...
	bitperioddelay( OneAndAHalfBitPeriod ) ;
	
	for ( uint8_t bitpos = 8 ; bitpos > 0 ; bitpos-- )
	{
		c = c >> 1 ;
		if (!SERIAL_PIN_MARK)
		{ // keep the 'do nothing' part at the top to keep the #instructions top and bottom the same.
			asm volatile ("nop");
		}
		else
		{
			c |= 0b10000000 ;
			BitCount++;
		}
		// the short BitPeriod delay is used here because this loop adds an overhead of about 10 instructions.
		bitperioddelay( BitPeriod_SlightlyShort ) ;
	}

	// the stop bit MUST be mark (high)
	uint8_t mystatus = (SERIAL_PIN_MARK) ? SERIAL_OK_bm : SERIAL_ERROR_FRAME_bm ;
	
	// mark the BREAK flag in the status byte if we've taken ages to find this byte.
	if (TIFR & (1<<OCF1A)) mystatus |= SERIAL_BREAK_bm ;
	*status = mystatus ;
	return c ;
}

#define BUFFER_LEN 16
void __attribute__((noreturn)) serial_eeprom_load( uint8_t *e, uint16_t datalength, uint16_t streamoffset ){
	// routine to suck data from the (repeating) serial data stream and place
	// it into eeprom. If the stream isn't long enough, zeros will be used.
	//
	// Whilst serial data is being listened to, LEDs 7/8 will twinkle
	// when data is seen on the sense line.
	//
	// Whilst autobauding, LEDs 1/2 will flicker.
	// When autobauding is complete, LED 3 will light.
	// When the load is complete, LED 5 will light and 7/8 will be stopped.
uint8_t data[BUFFER_LEN] ;
uint8_t c ;
uint8_t status ;
uint16_t left_to_write = datalength ;
uint16_t nextoffset = streamoffset ;

	serial_init() ;
	serial_autobaud() ;
	DDRB |= (1<<PB1) ; // set PB1 to output
	PORTB |= (1<<PB1) ; // set PB1 high (will light LED3)

	while( left_to_write != 0 ){
		// wait for the first character
		do{
			c = serial_getbyte( &status ) ;
		} while (!(status & SERIAL_BREAK_bm)) ;
		
		// we've just read the first character in the serial stream...
		uint16_t pos = 0 ;
		do {
			if (pos == nextoffset) break ;
			c = serial_getbyte( &status ) ;
			pos++ ;
		} while (!(status & SERIAL_BREAK_bm)) ;

		uint8_t index = 0 ;
		if (pos==nextoffset){
			// we're at the right point in the stream to read data...
			do{
				data[index++] = c ;
				c = serial_getbyte( &status ) ;
			} while (!(status & SERIAL_BREAK_bm) && index < BUFFER_LEN) ;
		}
		// fill any remainder of the buffer with zeroes
		while( index < BUFFER_LEN ){
			data[index++] = 0 ;
		}

		// data[] is now filled with data (or zeroes if the stream was too short)
		index = 0 ;
		while( (left_to_write != 0) && (index < BUFFER_LEN) ){
			eeprom_write_byte( e++, data[index++] ) ;
			left_to_write-- ;
		}
		nextoffset += BUFFER_LEN ;
	}

	// all the data was written.
	// set the LEDs to say 'FINISHED'
	DDRB |= (1<<PB3) ; // set PB3 to be an output.
	PORTB |= (1<<PB3) ; // set PB3 high (will light LED 5)
	
	serial_finish();

	while(1); 

}
