/*
*/

#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "random.h"
#include "led.h"
#include "sweep.h"
#include "reset.h"
#include "transmit.h"
#include "receive.h"
#include "serial.h"

#define MYLEDCOUNT 8
#define BASE_DELAY_TICKS 10

typedef void pattern_function ( void ) ;

/* pattern functions */
pattern_function cylonsweep ;
pattern_function sweep ;
pattern_function reversesweep ;
pattern_function brownian ;
pattern_function twinkle ;
pattern_function simplerandom ;
pattern_function stacking ;
pattern_function stacking_b ;
pattern_function userpattern ;


#define PATTERN_COUNT 8
pattern_function * PROGMEM f_pattern_function[ PATTERN_COUNT ] =
{
	cylonsweep,
	sweep,
	reversesweep,
	simplerandom,
	brownian,
	stacking,
	stacking_b,
	userpattern
} ;

typedef uint8_t bool ;
#define false 0
#define true 1

static uint8_t internal_random_getbyte(void);
void cmd( uint8_t cmd, uint8_t op1, uint8_t op2 ) ;
void tx( uint16_t data ) ;
void cmd_clearled( uint8_t ledindex ) ;
void cmd_setled( uint8_t ledindex ) ;
void cmd_setbinary( uint8_t index, uint8_t data ) ;
void cmd_userpattern( uint8_t index ) ;
void cmd_clearall( void ) ;
void cmd_setstyle( StyleID_t s ) ;

void common_delay( uint8_t ticks ){
	// how many ticks to delay for...

	// g_tickdelay holds how many 10th of a ms there are in a tick.
	// in medium mode, it might be 40, meaning a tick is 4ms.
	uint16_t count_tenths_ms ;
	count_tenths_ms = g_tickdelay * ticks ;
	for( uint16_t count = 0 ; count < count_tenths_ms ; count++ ){
		_delay_us( 100 ) ; // 1/10th ms
	}
}

void __attribute__((noreturn)) fn_sweep( uint8_t data ){
	tx_init() ;
	
	// save the current state of the data if we've just been reset...
	if (g_reset_reason & (1<<EXTRF)){
		g_function_data = data ;
		eeprom_save_state();
	}

	// now set up the next pattern - used if the button is pressed...
	uint8_t nextdata = data + 1 ;
	if (nextdata >= (PATTERN_COUNT * STYLE_COUNT)) nextdata = 0 ;
	g_function_data = nextdata ;

	// now calculate the rendering style and pattern index...
	uint8_t style = 0 ;
	uint8_t pattern = data ;
	while( pattern >= PATTERN_COUNT ){ pattern -= PATTERN_COUNT ; style++ ;}

	pattern_function *p_pattern_function = (pattern_function *)pgm_read_word( &f_pattern_function[ pattern ] ) ;
	sei() ;
	cmd_clearall() ;
	while(1){
		cmd_setstyle( style ) ;
		p_pattern_function() ;
	}
}

void __attribute__((noreturn)) fn_sweep_importpattern( void ){
	serial_eeprom_load( (uint8_t*)eeprom_get_sweepuserpattern(), 130, 0 ) ;
}

void cmd( uint8_t cmd, uint8_t op1, uint8_t op2 ){
	tx( ((cmd & 0b11111) << 11) | ((op1 & 0b111) << 8) | op2 ) ;
}

void tx( uint16_t data ){
	tx_send( data ) ;
	while(!tx_ready()) ; // wait until the data has been sent
	rx_insertdata( data ) ;
}

void cmd_clearled( uint8_t ledindex ){
	cmd( CMD_ClearLED,0,ledindex ) ;
}

void cmd_setled( uint8_t ledindex ){
	cmd( CMD_SetLED,0,ledindex ) ;
}

void cmd_setbinary( uint8_t index, uint8_t data ){
	cmd( CMD_SetBinary, index, data ) ;
}

void cmd_userpattern( uint8_t index ){
	cmd( CMD_UserPattern, 0, index ) ;
}

void cmd_clearall( void ){
	cmd( CMD_ClearAll, 0, 0 ) ;
}

void cmd_setstyle( StyleID_t s ){
	cmd( CMD_Style, 0, s) ;
}

void cmd_NULL( void ){
	cmd( CMD_NULL, 0, 0 ) ;
}

static uint8_t internal_random_getbyte( void ) ;

static uint8_t _lastled = 255 ;
void genericsweep( uint8_t start, uint8_t end ){
	if (end < start){
		// downwards sweep
		// assume that the previous LED has already been cleared...
		for( uint8_t i = (start + 1) ; i >= (end + 1) ; i-- ){
			cmd_clearled( _lastled ) ;
			cmd_setled( i - 1 ) ;
			_lastled = i - 1 ;
			common_delay( BASE_DELAY_TICKS ) ;
		}
	} else {
		// upwards sweep
		for( uint8_t i = start ; i <= end ; i++ ){
			cmd_clearled( _lastled ) ;
			cmd_setled( i ) ;
			_lastled = i ;
			common_delay( BASE_DELAY_TICKS ) ;
		}
	}
}
void nullsweep( uint8_t size ){
	// perform a sweep of length 'size' that does nothing except
	// take the same time as a normal sweep...
	// It's used to balance timing of some sweeps.
	for( uint8_t x = 0 ; x < size ; x++ ){
		cmd_clearled( _lastled ) ;
		cmd_NULL();
		common_delay( BASE_DELAY_TICKS ) ;
	}
}

void _stacking( bool regulartiming ){
	for( uint8_t i = 0 ; i < g_chain_ledlength ; i++ ){
		if (regulartiming) nullsweep(i);
		genericsweep( g_chain_ledlength - 1, i ) ;
		_lastled = 255 ;
	}
	for( uint8_t i = 0 ; i < g_chain_ledlength ; i++ ){
		genericsweep( i, 0 ) ;
		if (regulartiming) nullsweep( g_chain_ledlength - i ) ;
	}
}

void stacking( void ){
	_stacking( true ) ;
}

void stacking_b( void ){
	_stacking( false ) ;
}

void userpattern( void ){
	// read the first byte of the user pattern data.
	// it holds the number of bytes in the sequence.
	uint8_t *e ;
	e = eeprom_get_sweepuserpattern() ;
	uint8_t length = eeprom_read_byte( e ) ;
	uint8_t delayticks = eeprom_read_byte( e+1 ) ;
	for( uint8_t i = 0 ; i < length ; i++ ){
		cmd_userpattern( i ) ;
		common_delay( delayticks ) ;
	}
}

void sweep()
{
	genericsweep( 0, g_chain_ledlength - 1 ) ;
}

void reversesweep()
{
	genericsweep( g_chain_ledlength - 1, 0 ) ;
}

void cylonsweep()
{
	genericsweep( 0, g_chain_ledlength - 1 ) ;
	genericsweep( g_chain_ledlength - 2, 1 ) ;
}


void simplerandom()
{
	uint8_t display[8] ;
	for(uint8_t i = 0 ; i < g_chain_length ; i++ ){
		display[i] = internal_random_getbyte() ;
		cmd_setbinary( i, display[i] ) ;
	}
	while(1){
		uint8_t b = internal_random_getbyte() ;
		if ((b & 0b11000000) == 0b11000000)
		{
			uint8_t index = b & 0b111 ; // 0 to 7
			uint8_t shift = (b & 0b111000) >> 3 ;
			display[index] ^= 1 << shift ;
			cmd_setbinary( index, display[index] ) ;
		}
			common_delay( BASE_DELAY_TICKS ) ;
	}
	
}

#define BROWNIAN_MAX_DOTS 3
void brownian( void )
{
static uint8_t pos[ BROWNIAN_MAX_DOTS ];
static uint8_t dir[ BROWNIAN_MAX_DOTS ];
static uint8_t particleindex = BROWNIAN_MAX_DOTS;

	if (particleindex == BROWNIAN_MAX_DOTS){
		// this is the first call of the routine...
		//initialise the positions and directions of the dots
		for( uint8_t dotindex = 0 ; dotindex < BROWNIAN_MAX_DOTS ; dotindex++ )
		{
			pos[ dotindex ] = g_chain_ledlength >> 1 ; // start them all in the middle of the chain.
			dir[ dotindex ] = 1 ;
		}
		particleindex = 0 ;
	}

	// update the position of one of the particles...
	
	uint8_t oldpos = pos[ particleindex ] ;
	uint8_t r = (random_getbyte() & 0xff) ; 
	// update the direction of the indexed particle...
	if (r < 127){
		dir[ particleindex ] = 1 ;
	} else {
		dir[ particleindex ] = 2 ;
	}
	
	// update the position of the indexed particle...
	uint8_t newpos = oldpos ;
	if (dir[particleindex]==1){
		if (++newpos >= g_chain_ledlength){ newpos = 0 ; }
	} else {
		if (newpos){ newpos-- ; } else { newpos = g_chain_ledlength - 1 ; }
	}
	pos[particleindex] = newpos ;

	// find out if we need to clear the old position...
	uint8_t oldposused = 0 ;
	for( uint8_t i = 0 ; i < BROWNIAN_MAX_DOTS ; i++ ){
		if (pos[i] == oldpos){
			oldposused = 1 ;
			break ;
		}
	}
	
	if (oldposused == 0){
		// the old position isn't used by any other particles, so clear it.
		cmd_clearled( oldpos ) ;
	}
	
	cmd_setled( newpos ) ;
	
	particleindex += 1 ;
	if (particleindex >= BROWNIAN_MAX_DOTS) particleindex = 0 ;
	
	common_delay( BASE_DELAY_TICKS ) ;
}


static uint16_t	m_internal_randomSeed = 0x8765;
static uint8_t internal_random_getbyte(void)
{
   if((m_internal_randomSeed&0x8000)==0)
      m_internal_randomSeed=m_internal_randomSeed<<1;
   else {
      m_internal_randomSeed = m_internal_randomSeed<<1;
      m_internal_randomSeed = m_internal_randomSeed ^ 0x1D87;
      }
   return((m_internal_randomSeed >> 8) & 0xff);
} 
