/*
*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "globals.h"
#include "transmit.h"

volatile uint8_t _tx_bitsleft __attribute__ ((section (".noinit"))) ;
volatile uint8_t _tx_data_hi __attribute__ ((section (".noinit"))) ;
volatile uint8_t _tx_data_lo __attribute__ ((section (".noinit"))) ;

/*
Data is transmitted using PB2 (pin 7).
Only one device in a chain should ever transmit.

All the functions in this module follow g_transmit to determine if
data should actually be transmitted on the pin.
All timings should be the same, regardless of wether the pin is actually
toggled or not.

Data is sent as 16bits, MSB first, LSB last.
There is no parity/checksum data.
The line is 'idle high', e.g. the pin is usually held at Vcc.

Byte transmission:
	Start signal
	Bit 15
	Bit 14
	...
	Bit 0


Start signal: (>256us high/idle)
 ___________________________________  _Vcc
                                      _Gnd
  |------- 256 us (min)   --------|
            (min)

zero bit:
_              ________
 |____________|
 
 |--- 64us ---|- 32us -|
 |------- 96 us -------|


one bit:
_   ____________________
 |_|
 
 0.5us
 |-------- 96 us -------|


The transmission is therefore based on 32us chunks of time;

start: >8 high
zero: 2 low, 1 high
one: blip, 3 high

Timer1 is used.
OCR1A times a point to set output high.
OCR2A marks the end of the bit period.

The receiver will be configured to wait for a falling edge.
It'll then read the line between 1us and 64us after the falling edge.

If no bit is received for 256us, the next bit is a start bit.

Note: 32us is 256 CPU ticks at 8Mhz. 256us is 2048 CPU ticks at 8Mhz
*/

ISR( TIM1_COMPA_vect ){
	if (g_transmit == TRANSMIT_ENABLE) PORTB |= SENSE_bm ; // pin high.
}

ISR( TIM1_COMPB_vect ){
static uint8_t s_data ;
	if (_tx_bitsleft == 16) s_data = _tx_data_hi ;
	if (_tx_bitsleft == 8) s_data = _tx_data_lo ;
	
	if (_tx_bitsleft){
		if (g_transmit == TRANSMIT_ENABLE) PORTB &= ~(SENSE_bm) ;// pin low.
		_tx_bitsleft-- ;
		if (s_data & 0b10000000){
			// a 1 needs to be transmitted.
			asm("nop");
			if (g_transmit == TRANSMIT_ENABLE) PORTB |= SENSE_bm ; // pin high.
		} else {
			// a 0 needs to be transmitted.
			// leave the pin low.
		}
		s_data <<= 1 ;
		OCR1A = 2 ; // 64 us low
		OCR1B = 3 ; // 32 us high
		TCNT1 = 0 ;
	} else
	{
		// the byte has finished being transmitted.
		// clear the OCR1A and OCR1B interrupts...
		TIMSK &= ~((1<<OCIE1A)|(1<<OCIE1B)) ;
	}
}

void tx_init( void ){
	// initialise data...
	_tx_bitsleft = 0 ;
	_tx_data_hi = 0 ;
	_tx_data_lo = 0 ;

	if (g_transmit == TRANSMIT_ENABLE){
		// set the pin to be an output, and high...
		PORTB |= SENSE_bm ;
		DDRB |= SENSE_bm ;
	}
	
	// initialise timer 1
	// set to be /256, meaning each timer tick is 32us.
	TCCR1 |= (0b1001 << CS10) ;
}

uint8_t tx_ready( void ){
	if (TIMSK & (1<<OCIE1A)) return 0 ; // still transmitting.
	return 1 ;
}

void tx_send( uint16_t data ){
	while(!tx_ready()); // wait until the previous data's been sent.
	// set up the data...
	_tx_bitsleft = 16 ;
	_tx_data_hi = (data >> 8) & 0xff ;
	_tx_data_lo = data & 0xff ;

	// and now output the start signal...
	if (g_transmit == TRANSMIT_ENABLE){
		PORTB |= SENSE_bm ; // ensure the TX pin's high (it should be anyway)
	}
	OCR1A = 0 ; // no low periods.
	OCR1B = 12 ; // = 384us high.
	TCNT1 = 0 ;
	GTCCR |= (1<<PSR1) ; // reset the prescaler
	TIFR = (1<<OCF1A)|(1<<OCF1B) ; // clear any compare flags

	// now enable the timer interrupts...
	TIMSK |= (1<<OCIE1A)|(1<<OCIE1B) ;
}
